/**
 * tkz-euclide 在 js 中的模仿。注意函数名称和用法与 tkz-euclide 略有区别，并且函数不够完全。
 * @author: maven
 * @version: 1.2
 * @copyleft
 */

class Matrix {
	/**
	 * @brief 两个矩阵(向量)相乘
	 * @param A 矩阵(向量) A
	 * @param B 矩阵(向量) B
	 * @return 返回矩阵(向量) A 左乘矩阵(向量) B
	 */
	static DOT(A, B) {
		if (A.length == 3 && !Array.isArray(A[0]) && B.length == 3 && B[0].length == 3) {
			let C = [];
			for (let j = 0; j < 3; j++) {
				// A 是 1*3 的向量，B 是 3*3 的矩阵
				let sum = 0;
				for (let i = 0; i < 3; i++) {
					sum += B[i][j] * A[i];
				}
				C[j] = sum;
			}
			return C;
		}
		if (A.length == 3 && A[0].length == 3 && B.length == 3 && !Array.isArray(B[0])) {
			let C = [];
			for (let j = 0; j < 3; j++) {
				// B 是 3*1 的向量，A 是 3*3 的矩阵
				let sum = 0;
				for (let i = 0; i < 3; i++) {
					sum += A[j][i] * B[i];
				}
				C[j] = sum;
			}
			return C;
		}
		if (A.length == 3 && A[0].length == 3 && B.length == 3 && B[0].length == 3) {
			let C = [];
			for (let i = 0; i < 3; i++) {
				C[i] = [];
				for (let j = 0; j < 3; j++) {
					C[i][j] = 0;
					for (let k = 0; k < 3; k++)
						C[i][j] += A[i][k] * B[k][j];
				}
			}
			return C;
		}
	}
	/**
	 * @brief 求 A 的伴随矩阵
	 * @param A 输入矩阵
	 * @return 返回输入矩阵的伴随矩阵
	 */
	static ADJ(A) {
		if (A.length == 3 && A[0].length == 3) {
			let C = [];
			for (let i = 0; i < 3; i++) {
				C[i] = [];
				for (let j = 0; j < 3; j++) {
					let [m, n, p, q] = [(i + 1) % 3, (i + 2) % 3, (j + 1) % 3, (j + 2) % 3];
					C[i][j] = A[m][p] * A[n][q] - A[m][q] * A[n][p];
				}
			}
			return C;
		}
	}
	/**
	 * @brief 平移矩阵
	 * @param a1 平移方向向量的第一分量
	 * @param a2 平移方向向量的第二分量
	 * @return 返回平移矩阵
	 */
	static T(a1, a2) {
		return [[1, 0, a1], [0, 1, a2], [0, 0, 1]];
	}
	/**
	 * @brief 绕原点的旋转矩阵
	 * @param theta 旋转角，弧度制
	 * @return 返回旋转矩阵
	 */
	static R(theta) {
		return [
			[Math.cos(theta), -Math.sin(theta), 0],
			[Math.sin(theta), Math.cos(theta), 0],
			[0, 0, 1]
		];
	}
	/**
	 * @brief 矩阵转置
	 * @param A 输入矩阵
	 * @return 返回输入矩阵的转置矩阵
	 */
	static Transpose(A) {
		let C = [];
		for (let i = 0; i < 3; i++) {
			C[i] = [];
			for (let j = 0; j < 3; j++) {
				C[i][j] = A[j][i];
			}
		}
		return C;
	}
}
/**
 * 点
 */
class tkzPoint {
	/**
	 * @param name 点名称
	 * @param x 横坐标
	 * @param y 纵坐标
	 */
	constructor(name, x, y) {
		this.name = name;
		this.x = parseFloat(x);
		this.y = parseFloat(y);
	}
}
/**
 * 直线
 */
class tkzLine {
	/**
	 * @param x1 直线上点 1 的横坐标
	 * @param y1 直线上点 1 的纵坐标
	 * @param x2 直线上点 2 的横坐标
	 * @param y2 直线上点 2 的纵坐标
	 */
	constructor(x1, y1, x2, y2) {
		this.x1 = parseFloat(x1);
		this.y1 = parseFloat(y1);
		this.x2 = parseFloat(x2);
		this.y2 = parseFloat(y2);
	}
	/**
	 * @brief 取得直线的参数方程
	 * @param t 参数
	 * @return 返回参数方程
	 */
	getPE(t) {
		let theta = Math.atan((this.y2 - this.y1) / (this.x2 - this.x1));
		let x = this.x1 + t * Math.cos(theta);
		let y = this.y1 + t * Math.sin(theta);
		return { 'x': x, 'y': y };
	}
}
/**
 * 椭圆
 */
class tkzEllipse {
	/**
	 * @param x0 中心点横坐标
	 * @param y0 中心点纵坐标
	 * @param a 标准方程中的 a
	 * @param b 标准方程中的 b
	 * @param alpha 倾角，弧度制
	 */
	constructor(x0, y0, a, b, alpha) {
		this.x0 = parseFloat(x0);
		this.y0 = parseFloat(y0);
		this.a = parseFloat(a);
		this.b = parseFloat(b);
		this.alpha = parseFloat(alpha);
	}
	/**
	 * @brief 取得圆锥曲线的类型
	 * @return 返回圆锥曲线的类型
	 */
	getType() {
		return 'ellipse';
	}
	/**
	 * @brief 取得曲线的参数方程
	 * @param t 参数
	 * @return 返回参数方程
	 */
	getPE(t) {
		let v = [this.a * Math.cos(t), this.b * Math.sin(t), 1];
		let [x, y, z] = Matrix.DOT(Matrix.R(this.alpha), v);
		[x, y, z] = Matrix.DOT(Matrix.T(this.x0, this.y0), [x, y, z]);
		return { 'x': x, 'y': y };
	}
	/**
	 * @brief 取得顶点的坐标
	 * @return 返回四个顶点的坐标，前两个为 a 方向的顶点，后两个为 b 方向的顶点
	 */
	getVertex() {
		let [ax1, ay1, az1] = Matrix.DOT(Matrix.R(this.alpha), [this.a, 0, 1]);
		[ax1, ay1, az1] = Matrix.DOT(Matrix.T(this.x0, this.y0), [ax1, ay1, az1]);
		let [ax2, ay2, az2] = Matrix.DOT(Matrix.R(this.alpha), [-this.a, 0, 1]);
		[ax2, ay2, az2] = Matrix.DOT(Matrix.T(this.x0, this.y0), [ax2, ay2, az2]);
		let [bx1, by1, bz1] = Matrix.DOT(Matrix.R(this.alpha + Math.PI / 2), [this.b, 0, 1]);
		[bx1, by1, bz1] = Matrix.DOT(Matrix.T(this.x0, this.y0), [bx1, by1, bz1]);
		let [bx2, by2, bz2] = Matrix.DOT(Matrix.R(this.alpha + Math.PI / 2), [-this.b, 0, 1]);
		[bx2, by2, bz2] = Matrix.DOT(Matrix.T(this.x0, this.y0), [bx2, by2, bz2]);
		return [{ 'x': ax1, 'y': ay1 }, { 'x': ax2, 'y': ay2 }, { 'x': bx1, 'y': by1 }, { 'x': bx2, 'y': by2 }];
	}
	/**
	 * @brief 取得焦点的坐标
	 * @return 返回两个焦点的坐标
	 */
	getF() {
		let c = Math.sqrt(Math.abs(this.a * this.a - this.b * this.b));
		let [v1, v2] = (this.a > this.b) ? [[c, 0, 1], [-c, 0, 1]] : [[0, c, 1], [0, -c, 1]];
		let [fx1, fy1, fz1] = Matrix.DOT(Matrix.R(this.alpha), v1);
		[fx1, fy1, fz1] = Matrix.DOT(Matrix.T(this.x0, this.y0), [fx1, fy1, fz1]);
		let [fx2, fy2, fz2] = Matrix.DOT(Matrix.R(this.alpha), v2);
		[fx2, fy2, fz2] = Matrix.DOT(Matrix.T(this.x0, this.y0), [fx2, fy2, fz2]);
		return [{ 'x': fx1, 'y': fy1 }, { 'x': fx2, 'y': fy2 }];
	}
	/**
	 * @brief 取得曲线的矩阵
	 * @return 返回曲线的矩阵
	 */
	getMatrix() {
		let matrix = [
			[this.b * this.b, 0, 0],
			[0, this.a * this.a, 0],
			[0, 0, -this.a * this.a * this.b * this.b]
		];
		let r = Matrix.R(-this.alpha);
		let t = Matrix.T(-this.x0, -this.y0);
		let r_T = Matrix.Transpose(r);
		let t_T = Matrix.Transpose(t);
		let m = Matrix.DOT(t_T, r_T);
		m = Matrix.DOT(m, matrix);
		m = Matrix.DOT(m, r);
		m = Matrix.DOT(m, t);
		return m;
	}
}
/**
 * 双曲线
 */
class tkzHyperbola {
	/**
	 * @param x0 中心点横坐标
	 * @param y0 中心点纵坐标
	 * @param a 标准方程中的 a
	 * @param b 标准方程中的 b
	 * @param alpha 倾角，弧度制
	 */
	constructor(x0, y0, a, b, alpha) {
		this.x0 = parseFloat(x0);
		this.y0 = parseFloat(y0);
		this.a = parseFloat(a);
		this.b = parseFloat(b);
		this.alpha = parseFloat(alpha);
	}
	/**
	 * @brief 取得圆锥曲线的类型
	 * @return 返回圆锥曲线的类型
	 */
	getType() {
		return 'hyperbola';
	}
	/**
	 * @brief 取得曲线的参数方程
	 * @param t 参数
	 * @return 返回参数方程
	 */
	getPE(t) {
		let [a, b] = [Math.sign(this.a) * this.a, Math.sign(this.b) * this.b];
		let v = [a / Math.cos(t), b * Math.tan(t), 1];
		let [x, y, z] = Matrix.DOT(Matrix.R(this.alpha), v);
		[x, y, z] = Matrix.DOT(Matrix.T(this.x0, this.y0), [x, y, z]);
		return { 'x': x, 'y': y };
	}
	/**
	 * @brief 取得曲线的顶点坐标
	 * @return 返回曲线的四个顶点坐标，前两个为 a 方向的顶点，后两个为 b 方向的顶点
	 */
	getVertex() {
		let [ax1, ay1, az1] = Matrix.DOT(Matrix.R(this.alpha), [Math.abs(this.a), 0, 1]);
		[ax1, ay1, az1] = Matrix.DOT(Matrix.T(this.x0, this.y0), [ax1, ay1, az1]);
		let [ax2, ay2, az2] = Matrix.DOT(Matrix.R(this.alpha), [-Math.abs(this.a), 0, 1]);
		[ax2, ay2, az2] = Matrix.DOT(Matrix.T(this.x0, this.y0), [ax2, ay2, az2]);
		let [bx1, by1, bz1] = Matrix.DOT(Matrix.R(this.alpha + Math.PI / 2), [Math.abs(this.b), 0, 1]);
		[bx1, by1, bz1] = Matrix.DOT(Matrix.T(this.x0, this.y0), [bx1, by1, bz1]);
		let [bx2, by2, bz2] = Matrix.DOT(Matrix.R(this.alpha + Math.PI / 2), [-Math.abs(this.b), 0, 1]);
		[bx2, by2, bz2] = Matrix.DOT(Matrix.T(this.x0, this.y0), [bx2, by2, bz2]);
		let points = [{ 'x': ax1, 'y': ay1 }, { 'x': ax2, 'y': ay2 }, { 'x': bx1, 'y': by1 }, { 'x': bx2, 'y': by2 }];
		return points;
	}
	/**
	 * @brief 取得曲线焦点的坐标
	 * @return 返回曲线两个焦点的坐标
	 */
	getF() {
		let c = Math.sqrt(this.a * this.a + this.b * this.b);
		let [v1, v2] = [[c, 0, 1], [-c, 0, 1]];
		let [fx1, fy1, fz1] = Matrix.DOT(Matrix.R(this.alpha), v1);
		[fx1, fy1, fz1] = Matrix.DOT(Matrix.T(this.x0, this.y0), [fx1, fy1, fz1]);
		let [fx2, fy2, fz2] = Matrix.DOT(Matrix.R(this.alpha), v2);
		[fx2, fy2, fz2] = Matrix.DOT(Matrix.T(this.x0, this.y0), [fx2, fy2, fz2]);
		return [{ 'x': fx1, 'y': fy1 }, { 'x': fx2, 'y': fy2 }];
	}
	/**
	 * @brief 取得曲线的矩阵
	 * @return 返回曲线的矩阵
	 */
	getMatrix() {
		let matrix = [
			[this.b * this.b, 0, 0],
			[0, -this.a * this.a, 0],
			[0, 0, -this.a * this.a * this.b * this.b]
		];
		let r = Matrix.R(-this.alpha);
		let t = Matrix.T(-this.x0, -this.y0);
		let r_T = Matrix.Transpose(r);
		let t_T = Matrix.Transpose(t);
		let m = Matrix.DOT(t_T, r_T);
		m = Matrix.DOT(m, matrix);
		m = Matrix.DOT(m, r);
		m = Matrix.DOT(m, t);
		return m;
	}
}
/**
 * 抛物线
 */
class tkzParabola {
	/**
	 * @param x0 顶点横坐标
	 * @param y0 顶点纵坐标
	 * @param p 焦准距
	 * @param alpha 倾角，弧度制
	 */
	constructor(x0, y0, p, alpha) {
		this.x0 = parseFloat(x0);
		this.y0 = parseFloat(y0);
		this.p = parseFloat(p);
		this.alpha = parseFloat(alpha);
	}
	/**
	 * @brief 取得圆锥曲线的类型
	 * @return 返回圆锥曲线的类型
	 */
	getType() {
		return 'parabola';
	}
	/**
	 * @brief 取得参数方程
	 * @param t 参数
	 * @return 返回参数方程
	 */
	getPE(t) {
		let v = [2 * this.p / (Math.tan(t) * Math.tan(t)), 2 * this.p / Math.tan(t), 1];
		let [x, y, z] = Matrix.DOT(Matrix.R(this.alpha), v);
		[x, y, z] = Matrix.DOT(Matrix.T(this.x0, this.y0), [x, y, z]);
		return { 'x': x, 'y': y };
	}
	/**
	 * @brief 取得焦点的坐标
	 * @return 返回焦点的坐标
	 */
	getF() {
		let v = [this.p / 2, 0, 1];
		let [fx, fy, fz] = Matrix.DOT(Matrix.R(this.alpha), v);
		[fx, fy, fz] = Matrix.DOT(Matrix.T(this.x0, this.y0), [fx, fy, fz]);
		return [{ 'x': fx, 'y': fy }];
	}
	/**
	 * @brief 取得曲线的矩阵
	 * @return 返回曲线的矩阵
	 */
	getMatrix() {
		let matrix = [
			[0, 0, this.p],
			[0, -1, 0],
			[this.p, 0, 0]
		];
		let r = Matrix.R(-this.alpha);
		let t = Matrix.T(-this.x0, -this.y0);
		let r_T = Matrix.Transpose(r);
		let t_T = Matrix.Transpose(t);
		let m = Matrix.DOT(t_T, r_T);
		m = Matrix.DOT(m, matrix);
		m = Matrix.DOT(m, r);
		m = Matrix.DOT(m, t);
		return m;
	}
}

class UUID {
	static fun() {
		return (((1 + Math.random()) * 0x10000) | 0).toString(16).substring(1);
	}
	static GET() {
		return ('tkz_tmp_' + UUID.fun() + UUID.fun() + "-" + UUID.fun() + "-3" + UUID.fun().substr(0, 2) + "-" + UUID.fun() + "-" + UUID.fun() + UUID.fun() + UUID.fun()).toLowerCase();
	}
}

/**
 * @brief tkz 画板
 * @param ctx: canvers 对象
 */
function tkz(ctx) {
	let it = this;
	/**
	 * 所有点的集合，以点名称为索引
	 */
	this.points = {};
	/**
	 * 所有元素的集合，以元素名称为索引，其中存放椭圆、直线等元素
	 */
	this.elements = {};
	/**
	 * 临时元素栈
	 */
	this.tkz_tmp_elements = [];
	// pt 到 cm 的转换系数
	this.cm_pt = 37.8;
	/**
	 * @brief 初始化 ctx。
	 * @param x 原点在画布上的 x 坐标
	 * @param y 原点在画布上的 y 坐标
	 */
	this.init = function (x, y) {
		ctx.translate(x, y);
		// 使 y 轴向上
		ctx.scale(1, -1);
	}
	/**
	 * @brief 设置填充颜色、笔触颜色、笔画样式等参数
	 * @param params 绘制参数
	 * @return 返回设置后的绘制参数
	 */
	function set_params(params) {
		params.color = params.color || 'black';
		params.fill = params.fill || 'black';
		params.width = params.width || 1;

		if (params.dash) {
			ctx.setLineDash([5, 5]);
		} else {
			ctx.setLineDash([]);
		}
		params.font = params.font || "14px Arial";
		return params;
	}

	/**
	 * @brief 判断数值是否为零
	 * @param number 被判断的数值
	 * @return 数值为零返回 true，否则返回 false
	 */
	function is_zero(number) {
		return Math.abs(number) < 1E-8;
	}
	/**
	 * @brief 过两点的直线的一般式，返回 Ax+By+C = 0 中的 A,B,C
	 * @param start 起点
	 * @param to 终点
	 * @return 返回 Ax+By+C = 0 中的 A,B,C
	 */
	function getABC(start, to) {
		return { A: (to.y - start.y), B: (start.x - to.x), C: (to.x * start.y - start.x * to.y) };
	}

	/**
	 * @brief 取得圆半径
	 * @param center 圆心
	 * @param point 圆上一点
	 * @return 返回圆半径
	 */
	function getR(center, point) {
		return Math.sqrt((center.x - point.x) * (center.x - point.x) + (center.y - point.y) * (center.y - point.y));
	}
	/**
	 * @brief 得到过两点直线的倾角，返回倾角
	 * @param start 起点
	 * @param to 终点
	 * @return 返回倾角
	 */
	function getDip(start, to) {
		let d = to.x - start.x;
		return (is_zero(d) ? 90 : Math.atan((to.y - start.y) / (to.x - start.x)) * 180 / Math.PI);
	}
	/**
	 * @brief 判断点是否在直线上
	 * @param one 直线上一点
	 * @param two 直线上另一点
	 * @param point 被判断的点
	 * @return 在直线上返回 true，否则返回 false
	 */
	function at_line(one, two, point) {
		let abc = getABC(one, two);
		return is_zero(abc.A * point.x + abc.B * point.y + abc.C);
	}
	/**
	 * @brief 判断点是否在圆周上
	 * @param center 圆心
	 * @param point 圆周上一点
	 * @param p 被判断的点
	 */
	function at_circle(center, point, p) {
		let r2 = (center.x - point.x) * (center.x - point.x) + (center.y - point.y) * (center.y - point.y);
		let d2 = (center.x - p.x) * (center.x - p.x) + (center.y - p.y) * (center.y - p.y);
		return is_zero(r2 - d2);
	}
	/**
	 * @brief 内部方法，创建一些点。用于其它方法取得交点后的链式调用，点坐标由临时变量取得
	 * @param points_name 点序列名称，用逗号分隔，例如 'A,B'，将从内部创建 A,B 两个点
	 * @return 返回被创建的所有点的数组，如果数组中只有一个点，返回这个点
	 */
	function create_point(points_name) {
		points_name = points_name.split(',');
		let points = [];
		for (let i = 0; i < points_name.length; i++) {
			let point_name = points_name[i];
			let xy = it.tkz_tmp_points[i];
			let point = new tkzPoint(point_name, xy.x, xy.y);
			if (!point.name.startsWith('tkz_tmp_')) {
				console.log(point);
			}
			it.points[point_name] = point;
			points.push(point);
		}
		return ((points.length > 1) ? points : points[0]);
	}
	/**
	 * @brief 创建点失败时调用此函数
	 * @param point_name 创建失败的点名称
	 */
	function create_point_err(point_name) {
		console.log(`点 ${point_name} 不能被创建: ${it.tkz_msg}`);
	}
	/**
	 * @brief 取得两点间的距离
	 * @param one_name 第一点
	 * @param two_name 第二点
	 * @param dir 距离是否有向，坐标轴正方向为正向。默认为无向距离，当值为 true 时为有向距离
	 * @return 返回两点间的距离
	 */
	this.tkzCalcLength = function (one_name, two_name, dir) {
		let one = this.points[one_name];
		let two = this.points[two_name];
		// 无向距离
		let d = Math.sqrt((one.x - two.x) * (one.x - two.x) + (one.y - two.y) * (one.y - two.y));
		if (dir) {
			if (is_zero(one.x - two.x)) {
				// x 坐标相等
				d *= (two.y < one.y ? -1 : 1);
			} else {
				// x 坐标不等
				d *= (two.x < one.x ? -1 : 1);
			}
		}
		return d;
	}
	/**
	 * @brief 定义点
	 * @param points_strs 点的字符串，例如 '1/0/A,2/0/B' 将定义 A(1,0) 和 B(2,0) 两个点
	 */
	this.tkzDefPoints = function (points_strs) {
		let point_strs = points_strs.split(',');
		for (let point_str of point_strs) {
			let point_params = point_str.split('/');
			this.tkz_tmp_points = [{ 'x': point_params[0], 'y': point_params[1] }];
			create_point(point_params[2]);
		}
	}
	/**
	 * @brief 创建两点的中点。使用方法：tkzDefMidPoint('A','B')('Y')，创建线段 AB 的中点 Y
	 * @param one_name 第一点
	 * @param tow_name 第二点
	 * @return 返回创建点函数，创建一个点。
	 */
	this.tkzDefMidPoint = function (one_name, tow_name) {
		let one = this.points[one_name];
		let two = this.points[tow_name];
		let x = (one.x + two.x) / 2;
		let y = (one.y + two.y) / 2;
		this.tkz_tmp_points = [{ 'x': x, 'y': y }];
		return create_point;
	}
	/**
	 * @brief 旋转点。使用方法：tkzDefRotaPoint('O','B',-45)('Z')，将点 B 绕点 O 顺时针旋转 45 度，得到点 Z。
	 * @param center_name 旋转中心
	 * @param point_name 被旋转的点
	 * @param angle 旋转角
	 * @return 返回创建点函数，创建一个点。
	 */
	this.tkzDefRotaPoint = function (center_name, point_name, angle) {
		let center = this.points[center_name];
		let point = this.points[point_name];
		let theta = -angle * Math.PI / 180;
		let x = (point.x - center.x) * Math.cos(theta) + (point.y - center.y) * Math.sin(theta) + center.x;
		let y = -(point.x - center.x) * Math.sin(theta) + (point.y - center.y) * Math.cos(theta) + center.y;
		this.tkz_tmp_points = [{ 'x': x, 'y': y }];
		return create_point;
	}
	/**
	 * @brief 旋转并缩放点。使用方法：tkzDefRotaAndScalePoint('O','B',-45, 2)('Z')，将点 B 绕点 O 顺时针旋转 45 度得到点 X，再按 OX 方向缩放点 X 到点 Z，使得 OZ:ZX = K。
	 * @param center_name 旋转中心
	 * @param point_name 被旋转的点
	 * @param angle 旋转角
	 * @param K 缩放定比
	 * @return 返回创建点函数，创建一个点。
	 */
	this.tkzDefRotaAndScalePoint = function (center_name, point_name, angle, K) {
		let X_name = UUID.GET();
		this.tkzDefRotaPoint(center_name, point_name, angle)(X_name);
		return this.tkzDefLinearPoint(center_name, X_name, K);
	}
	/**
	 * @brief 过一点作直线的垂线，得到垂足点。使用方法：tkzDefProjPoint('A','B','C')('Z')，过 C 作 AB 的垂线，得到垂足点 Z。
	 * @param one_name 直线上一点
	 * @param two_name 直线上另一点
	 * @param point_name 直线外一点
	 * @return 返回创建点函数，创建一个点。
	 */
	this.tkzDefProjPoint = function (one_name, two_name, point_name) {
		let one = this.points[one_name];
		let two = this.points[two_name];
		let point = this.points[point_name];
		let abc = getABC(one, two);

		let d = abc.A * abc.A + abc.B * abc.B;
		let x = (abc.B * abc.B * point.x - abc.A * abc.B * point.y - abc.A * abc.C) / d;
		let y = (abc.A * abc.A * point.y - abc.A * abc.B * point.x - abc.B * abc.C) / d;
		this.tkz_tmp_points = [{ 'x': x, 'y': y }];
		return create_point;
	}
	/**
	 * @brief 作一点关于另两点连线的对称点，得到对称点。使用方法：tkzDefReflPoint('A','B','C')(`C'`)，作 C 关于 AB 的对称点C'。
	 * @param one_name 直线上一点
	 * @param two_name 直线上另一点
	 * @param point_name 被对称的点
	 * @return 返回创建点函数，创建一个点。
	 */
	this.tkzDefReflPoint = function (one_name, two_name, point_name) {
		let one = this.points[one_name];
		let two = this.points[two_name];
		let point = this.points[point_name];
		let abc = getABC(one, two);
		let d = abc.A * abc.A + abc.B * abc.B;
		let x = ((abc.B * abc.B - abc.A * abc.A) * point.x - 2 * abc.A * (abc.B * point.y + abc.C)) / d;
		let y = ((abc.A * abc.A - abc.B * abc.B) * point.y - 2 * abc.B * (abc.A * point.x + abc.C)) / d;
		this.tkz_tmp_points = [{ 'x': x, 'y': y }];
		return create_point;
	}
	/**
	 * @brief 定比分点。使用方法：tkzDefLinearPoint('A','B',2)('C')，定义点 C，使得 AC : CB = K。
	 * @param one_name 直线上一点
	 * @param two_name 直线上另一点
	 * @param K 定比
	 * @return 返回创建点函数，创建一个点。
	 */
	this.tkzDefLinearPoint = function (one_name, two_name, K) {
		if (is_zero(K + 1)) {
			// 禁止定比为 -1
			this.tkz_msg = '禁止定比为 -1';
			return create_point_err;
		}
		let one = this.points[one_name];
		let two = this.points[two_name];
		let x = (one.x + K * two.x) / (1 + K);
		let y = (one.y + K * two.y) / (1 + K);
		this.tkz_tmp_points = [{ 'x': x, 'y': y }];
		return create_point;
	}

	/**
	 * @brief 过一点作直线的垂线，并在垂线上取一点。使用方法：tkzDefOrthLine('A','B','P')('C')，过 P 作 AB 的垂线，并在垂线上取一点 C。
	 * @param one_name 直线上一点
	 * @param two_name 直线上另一点
	 * @param point_name 作垂线所过的点
	 * @return 返回创建点函数，创建一个点。
	 */
	this.tkzDefOrthLine = function (one_name, two_name, point_name) {
		let one = this.points[one_name];
		let two = this.points[two_name];
		let point = this.points[point_name];
		if (!at_line(one, two, point)) {
			// 点 point 不在直线上，取得点到直线的垂足
			return this.tkzDefProjPoint(one_name, two_name, point_name);
		} else {
			// 点 point 在直线上
			if (is_zero(point.x - one.x) && is_zero(point.y - one.y)) {
				// 点 point 就是点 one，将点 two 绕 point 旋转 90 度
				return this.tkzDefRotaPoint(point_name, two_name, 90);
			} else {
				// 点 point 不是点 one，将点 one 绕 point 旋转 90 度
				return this.tkzDefRotaPoint(point_name, one_name, 90);
			}
		}
	}
	/**
	 * @brief 过一点作直线的平行线，并在平行线上取一点。使用方法：tkzDefParaLine('A','B','P')('C')，过 P 作 AB 的平行线，并在平行线上取一点 C。
	 * @param one_name 直线上一点
	 * @param two_name 直线上另一点
	 * @param point_name 作平行线所过的点
	 * @return 返回创建点函数，创建一个点。
	 */
	this.tkzDefParaLine = function (one_name, two_name, point_name) {
		let one = this.points[one_name];
		let point = this.points[point_name];
		// 将点 two 绕点 one 旋转 90 度，得到点 X
		let X_name = UUID.GET();
		let X = this.tkzDefRotaPoint(one_name, two_name, 90)(X_name);
		if (!at_line(one, X, point)) {
			// 点 point 不在直线 one-X 上，取得点到直线的垂足
			return this.tkzDefProjPoint(one_name, X_name, point_name);
		} else {
			// 点 point 在直线 one-X 上
			// 将点 one 绕点 two 旋转 90 度，得到点 X
			this.tkzDefRotaPoint(two_name, one_name, 90)(X_name);
			// 取得点到直线的垂足
			return this.tkzDefProjPoint(two_name, X_name, point_name);
		}
	}
	/**
	 * @brief 求直线 2 到直线 1 的角。使用方法：tkzGetAngle('P1','P2','T1','T2')，得到 T1T2 到 P1P2 的角
	 * @param P1 直线 1 上第 1 点的名称
	 * @param P2 直线 1 上第 2 点的名称
	 * @param T1 直线 2 上第 1 点的名称
	 * @param T2 直线 2 上第 2 点的名称
	 * @return 返回直线 2 到直线 1 的角
	 */
	this.tkzGetAngle = function (P1, P2, T1, T2) {
		let l1_abc = getABC(this.points[P1], this.points[P2]);
		let l2_abc = getABC(this.points[T1], this.points[T2]);
		// 到角公式求 T1T2 到 P1P2 的角
		let d = l1_abc.A * l2_abc.A + l1_abc.B * l2_abc.B;
		let n = l1_abc.A * l2_abc.B - l2_abc.A * l1_abc.B;
		// 两直线互相垂直，到角为 90 度或 -90 度
		let theta = (n < 0) ? -90 : 90;
		// 两直线不互相垂直，计算到角
		theta = is_zero(d) ? theta : Math.atan(n / d) * 180 / Math.PI;
		return theta;
	}
	/**
	 * @brief 定义角平分线上一点。使用方法：tkzDefBisectorLine('A','B','C')('P')，在角 ABC 的平分线上取一点 P
	 * @param A 三角形的顶点 A
	 * @param B 三角形的顶点 B
	 * @param C 三角形的顶点 C
	 * @return 返回创建点函数，创建一个点。
	 */
	this.tkzDefBisectorLine = function (A, B, C) {
		// 取得到角
		let theta = this.tkzGetAngle(B, A, B, C);
		// 将 A 绕 B 旋转 theta/2
		return this.tkzDefRotaPoint(B, A, theta / 2);
	}

	/**
	 * @brief 取得两条直线的交点。使用方法：tkzInterLL('A,B,C,D')('X')，创建直线 AB,CD 的交点 X
	 * @param points_name 分别为两条直线上的两点名称，例如 'A,B,C,D' 表示直线 AB, CD 的交点。
	 * @return 返回创建点函数，创建一个点。
	 */
	this.tkzInterLL = function (points_name) {
		let point_names = points_name.split(',');
		let A1 = this.points[point_names[0]];
		let A2 = this.points[point_names[1]];
		let B1 = this.points[point_names[2]];
		let B2 = this.points[point_names[3]];
		let d = (A2.y - A1.y) * (B2.x - B1.x) - (A2.x - A1.x) * (B2.y - B1.y);
		let x = Infinity;
		let y = Infinity;
		if (!is_zero(d)) {
			// 两直线不平行，有交点
			let m1 = B1.y * B2.x - B1.x * B2.y;
			let m2 = A1.y * A2.x - A1.x * A2.y;

			x = ((B1.x - B2.x) * m2 - (A1.x - A2.x) * m1) / d;
			y = ((B1.y - B2.y) * m2 - (A1.y - A2.y) * m1) / d;
		}
		this.tkz_tmp_points = [{ 'x': x, 'y': y }];
		return create_point;
	}
	/**
	 * @brief 定义三角形的垂心。
	 * 使用方法：
	 * 	tkzOrthoCenter('A','B','C')('H')，创建三角形 ABC 的垂心 H
	 * 	tkzOrthoCenter('A','B','C')('H,H_{a},H_{b},H_{c}')，创建三角形 ABC 的垂心 H，以及垂心在三边的垂足 H_a, H_b, H_c
	 * @param A 三角形的顶点 A
	 * @param B 三角形的顶点 B
	 * @param C 三角形的顶点 C
	 * @return 返回创建点函数，创建一个点。
	 */
	this.tkzOrthoCenter = function (A, B, C) {
		let [Ha_name, Hb_name, Hc_name] = [UUID.GET(), UUID.GET(), UUID.GET()];
		let ha = this.tkzDefProjPoint(B, C, A)(Ha_name);
		let hb = this.tkzDefProjPoint(C, A, B)(Hb_name);
		let hc = this.tkzDefProjPoint(A, B, C)(Hc_name);
		let h = this.tkzInterLL(`${A},${Ha_name},${C},${Hc_name}`)(UUID.GET());
		this.tkz_tmp_points = [{ 'x': h.x, 'y': h.y }, { 'x': ha.x, 'y': ha.y }, { 'x': hb.x, 'y': hb.y }, { 'x': hc.x, 'y': hc.y }];
		return create_point;
	}
	/**
	 * @brief 定义三角形的外心。
	 * 使用方法：
	 * 	tkzCircumCenter('A','B','C')('O')，创建三角形 ABC 的外心 O
	 * 	tkzCircumCenter('A','B','C')('O,M_{a},M_{b},M_{c}')，创建三角形 ABC 的外心 O，以及三边中点 M_a, M_b, M_c
	 * @param A 三角形的顶点 A
	 * @param B 三角形的顶点 B
	 * @param C 三角形的顶点 C
	 * @return 返回创建点函数，创建一个点。
	 */
	this.tkzCircumCenter = function (A, B, C) {
		let [Ma_name, Mb_name, Mc_name] = [UUID.GET(), UUID.GET(), UUID.GET()];
		let [Xa_name, Xc_name] = [UUID.GET(), UUID.GET()];
		let O_name = UUID.GET();

		let ma = this.tkzDefMidPoint(B, C)(Ma_name);
		let mb = this.tkzDefMidPoint(C, A)(Mb_name);
		let mc = this.tkzDefMidPoint(A, B)(Mc_name);

		this.tkzDefOrthLine(A, B, Mc_name)(Xc_name);
		this.tkzDefOrthLine(B, C, Ma_name)(Xa_name);

		let o = this.tkzInterLL(`${Ma_name},${Xa_name},${Mc_name},${Xc_name}`)(O_name);

		this.tkz_tmp_points = [{ 'x': o.x, 'y': o.y }, { 'x': ma.x, 'y': ma.y }, { 'x': mb.x, 'y': mb.y }, { 'x': mc.x, 'y': mc.y }];
		return create_point;
	}
	/**
	 * @brief 定义三角形的内心。
	 * 使用方法：
	 * 	tkzInCenter('A','B','C')('I')，创建三角形 ABC 的内心 I
	 * 	tkzInCenter('A','B','C')('I,I_{a},I_{b},I_{c}')，创建三角形 ABC 的内心 I，以及内切圆在三边上的切点 I_a, I_b, I_c
	 * @param A 三角形的顶点 A
	 * @param B 三角形的顶点 B
	 * @param C 三角形的顶点 C
	 * @return 返回创建点函数，创建一个点。
	 */
	this.tkzInCenter = function (A, B, C) {
		let [Xb_name, Xc_name] = [UUID.GET(), UUID.GET()];
		this.tkzDefBisectorLine(A, B, C)(Xb_name);
		this.tkzDefBisectorLine(B, C, A)(Xc_name);

		let I_name = UUID.GET();
		let i = this.tkzInterLL(`${B},${Xb_name},${C},${Xc_name}`)(I_name);
		let ic = this.tkzDefProjPoint(A, B, I_name)(UUID.GET());
		let ia = this.tkzDefProjPoint(B, C, I_name)(UUID.GET());
		let ib = this.tkzDefProjPoint(C, A, I_name)(UUID.GET());
		this.tkz_tmp_points = [{ 'x': i.x, 'y': i.y }, { 'x': ia.x, 'y': ia.y }, { 'x': ib.x, 'y': ib.y }, { 'x': ic.x, 'y': ic.y }];
		return create_point;
	}
	/**
	 * @brief 定义三角形的重心。
	 * 使用方法：
	 * 	tkzBaryCenter('A','B','C')('G')，创建三角形 ABC 的重心 G
	 * 	tkzBaryCenter('A','B','C')('G,G_{a},G_{b},G_{c}')，创建三角形 ABC 的重心 G，以及三边中点 G_a, G_b, G_c
	 * @param A 三角形的顶点 A
	 * @param B 三角形的顶点 B
	 * @param C 三角形的顶点 C
	 * @return 返回创建点函数，创建一个点。
	 */
	this.tkzBaryCenter = function (A, B, C) {
		let [Xa_name, Xb_name, Xc_name] = [UUID.GET(), UUID.GET(), UUID.GET()];
		let ga = this.tkzDefMidPoint(B, C)(Xa_name);
		let gb = this.tkzDefMidPoint(C, A)(Xb_name);
		let gc = this.tkzDefMidPoint(A, B)(Xc_name);
		let g = this.tkzInterLL(`${A},${Xa_name},${B},${Xb_name}`)(UUID.GET());
		this.tkz_tmp_points = [{ 'x': g.x, 'y': g.y }, { 'x': ga.x, 'y': ga.y }, { 'x': gb.x, 'y': gb.y }, { 'x': gc.x, 'y': gc.y }];
		return create_point;
	}
	/**
	 * @brief 定义三角形的旁心。
	 * 使用方法：
	 * 	tkzExCenter('A','B','C')('E')，创建三角形 ABC 关于角 B 的旁心 E
	 * 	tkzExCenter('A','B','C')('E,E_{a},E_{b},E_{c}')，创建三角形 ABC 关于角 B 的旁心 E，以及旁切圆在三边的切点 E_a, E_b, E_c
	 * @param A 三角形的顶点 A
	 * @param B 三角形的顶点 B
	 * @param C 三角形的顶点 C
	 * @return 返回创建点函数，创建一个点。
	 */
	this.tkzExCenter = function (A, B, C) {
		let [Xb_name, Xc_name] = [UUID.GET(), UUID.GET()];
		let X_name = UUID.GET();

		this.tkzDefBisectorLine(A, B, C)(Xb_name);
		this.tkzDefBisectorLine(B, C, A)(Xc_name);
		this.tkzDefRotaPoint(C, Xc_name, 90)(X_name);

		let E_name = UUID.GET();
		let e = this.tkzInterLL(`${C},${X_name},${B},${Xb_name}`)(E_name);
		let ec = this.tkzDefProjPoint(A, B, E_name)(UUID.GET());
		let ea = this.tkzDefProjPoint(B, C, E_name)(UUID.GET());
		let eb = this.tkzDefProjPoint(C, A, E_name)(UUID.GET());
		this.tkz_tmp_points = [{ 'x': e.x, 'y': e.y }, { 'x': ea.x, 'y': ea.y }, { 'x': eb.x, 'y': eb.y }, { 'x': ec.x, 'y': ec.y }];
		return create_point;
	}
	/**
	 * @brief 取得直线与圆的交点。使用方法：tkzInterLC('A,B,O,P')('X,Y')，创建直线 AB 与以 O 为圆心 OP 为半径的圆的交点 X,Y。
	 * @param points_name 直线上两点的名称和圆心名称和圆上一点的名称，例如 'A,B,O,P' 表示直线 AB, 圆 OP。
	 * @return 返回创建点函数，创建两个点。
	 */
	this.tkzInterLC = function (points_name) {
		let point_names = points_name.split(',');
		let pA = this.points[point_names[0]];
		let pB = this.points[point_names[1]];
		let O = this.points[point_names[2]];
		let P = this.points[point_names[3]];
		// 圆半径
		let r = getR(O, P);
		// 直线一般方程的 A,B,C
		let abc = getABC(pA, pB);
		let m = abc.A * abc.A + abc.B * abc.B;

		let a2 = abc.A * abc.A;
		let b2 = abc.B * abc.B;
		let s1 = abc.C + abc.A * O.x;
		let s2 = abc.C + abc.B * O.y;

		let t = Math.sqrt(b2 * (m * r * r - (s1 + abc.B * O.y) * (s1 + abc.B * O.y)));
		if (!is_zero(abc.B)) {
			let x1 = (b2 * O.x - abc.A * s2 - t) / (a2 + b2);
			let y1 = (-b2 * s1 + a2 * abc.B * O.y + abc.A * t) / (abc.B * (a2 + b2));
			let x2 = (b2 * O.x - abc.A * s2 + t) / (a2 + b2);
			let y2 = (-b2 * s1 + a2 * abc.B * O.y - abc.A * t) / (abc.B * (a2 + b2));
			this.tkz_tmp_points = [{ 'x': x1, 'y': y1 }, { 'x': x2, 'y': y2 }];
		} else {
			let x = -abc.C / abc.A;
			let p = Math.sqrt(-a2 * s1 * s1 + a2 * a2 * r * r) / a2;
			let y1 = O.y - p;
			let y2 = O.y + p;
			this.tkz_tmp_points = [{ 'x': x, 'y': y1 }, { 'x': x, 'y': y2 }];
		}
		return create_point;
	}
	/**
	 * @brief 取得两圆的交点。使用方法：tkzInterCC('O1,A,O2,B')('X,Y')，创建圆 O1A 与圆 O2B 的交点 X,Y。
	 * @param points_name 圆心名称和圆上一点的名称对，例如 'O1,A,O2,B' 表示圆 O1A 与圆 O2B。
	 * @return 返回创建点函数，创建两个点。
	 */
	this.tkzInterCC = function (points_name) {
		let point_names = points_name.split(',');
		let O1 = this.points[point_names[0]];
		let A = this.points[point_names[1]];
		let O2 = this.points[point_names[2]];
		let B = this.points[point_names[3]];
		// 连心线长度
		let l = this.tkzCalcLength(O1.name, O2.name);
		// 圆 O1A 的半径
		let r1 = getR(O1, A);
		// 圆 O2B 的半径
		let r2 = getR(O2, B);
		// 设相交弦 P1P2 与连心线交于 M
		// 有 MP1^2 = r1^2-O1M^2 = r2^2-O2M^2，所以 r1^2-r2^2 = (O1M+O2M)(O1M-O2M) = l(O1M-O2M)
		// 计算 O1M 的长度
		let o1m = ((r1 * r1 - r2 * r2) / l + l) / 2;
		// 角 P1O1M
		let theta = Math.acos(o1m / r1) * 180 / Math.PI;

		let [x1, x2, y1, y2] = [0, 0, 0, 0];

		// 将 O2 绕 O1 旋转 theta，得到点 X
		let X_name = UUID.GET();
		this.tkzDefRotaPoint(O1.name, O2.name, theta)(X_name);
		//直线 O1X 与圆 O2B 的交点 Y1, Y2
		let [Y1, Y2] = this.tkzInterLC(`${O1.name},${X_name},${O2.name},${B.name}`)(`${UUID.GET()},${UUID.GET()}`);
		if (at_circle(O1, A, Y1) && at_circle(O2, B, Y1)) {
			// Y1 在两个圆上，是交点
			x1 = Y1.x;
			y1 = Y1.y;
		} else {
			// Y2 在两个圆上，是交点
			x1 = Y2.x;
			y1 = Y2.y;
		}
		// 将 O2 绕 O1 旋转 -theta，得到点 X
		this.tkzDefRotaPoint(O1.name, O2.name, -theta)(X_name);
		//直线 O1X 与圆 O2B 的交点
		let [Z1, Z2] = this.tkzInterLC(`${O1.name},${X_name},${O2.name},${B.name}`)(`${UUID.GET()},${UUID.GET()}`);
		if (at_circle(O1, A, Z1) && at_circle(O2, B, Z1)) {
			// Z1 在两个圆上，是交点
			x2 = Z1.x;
			y2 = Z1.y;
		} else {
			// Z2 在两个圆上，是交点
			x2 = Z2.x;
			y2 = Z2.y;
		}
		this.tkz_tmp_points = [{ 'x': x1, 'y': y1 }, { 'x': x2, 'y': y2 }];
		return create_point;
	}
	/**
	 * @brief 过圆外一点作圆的切线，取得两个切点。使用方法：tkzDefTangentFrom('O', 'A', 'P')('X,Y')，过 P 作圆 OA 的切线，取得切点 X,Y。
	 * @param center_name 圆心点名称
	 * @param point_name 圆上一点名称
	 * @param p_name 圆外一点名称
	 * @return 返回创建点函数，创建两个点。
	 */
	this.tkzDefTangentFrom = function (center_name, point_name, p_name) {
		return this.tkzInterCC(`${center_name},${point_name},${p_name},${center_name}`);
	}
	/**
	 * @brief 过圆上一点作圆的切线，并取得切线上一点。使用方法：tkzDefTangentAt('O', 'A')('X')，过 A 作圆 OA 的切线，取得切线上一点 X。
	 * @param center_name 圆心点名称
	 * @param point_name 圆上一点名称
	 * @return 返回创建点函数，创建一个点。
	 */
	this.tkzDefTangentAt = function (center_name, point_name) {
		return this.tkzDefOrthLine(center_name, point_name, point_name);
	}
	/**
	 * @brief 作两圆的内公切线，并取得切点。使用方法：tkzDefInTanCC('O1', 'P1', 'O2', 'P2')('X1,X2,Y1,Y2')，作圆 O1P1 和圆 O2P2 的内公切线 X1X2, Y1Y2，切点为 X1, X2, Y1, Y2
	 * @param O1_name 圆心 1 的名称
	 * @param P1_name 圆 1 上一点名称
	 * @param O2_name 圆心 2 的名称
	 * @param P2_name 圆 2 上一点名称
	 * @return 返回创建点函数，创建四个点。
	 */
	this.tkzDefInTanCC = function (O1_name, P1_name, O2_name, P2_name) {
		let O1 = this.points[O1_name];
		let O2 = this.points[O2_name];
		let P1 = this.points[P1_name];
		let P2 = this.points[P2_name];

		let O1O2 = this.tkzCalcLength(O1_name, O2_name);
		let r1 = getR(O1, P1);
		let r2 = getR(O2, P2);
		if (r1 + r2 - O1O2 > 0) {
			// 两圆相交或内含，没有内公切线
			this.tkz_msg = '两圆相交或内含，没有内公切线';
			return create_point_err;
		}
		// 设连心线与切线交点为 X，则有向比 O1X:XO2 = r1:r2
		let X_name = UUID.GET();
		let O1X = this.tkzCalcLength(O1_name, X_name);
		let theta = Math.acos(r1 / O1X) * 180 / Math.PI;
		// 将 O1X 绕 O1 转 theta 到 Y，并缩放到 Y1, Y2，使得 K = O1Y1:Y1Y = r1:(O1X-r1)
		let K1 = r1 / (O1X - r1);

		let [Y1_name, Y2_name] = [UUID.GET(), UUID.GET()];
		let Y1 = this.tkzDefRotaAndScalePoint(O1_name, X_name, theta, K1)(Y1_name);
		let Y2 = this.tkzDefRotaAndScalePoint(O1_name, X_name, -theta, K1)(Y2_name);

		// 将 O2X 绕 O2 转 theta 到 Z，并缩放到 Z1, Z2，使得 K = O2Z1:Z1Y = r2:(O2X-r2)
		let [Z1_name, Z2_name] = [UUID.GET(), UUID.GET()];
		let Z1 = this.tkzDefRotaAndScalePoint(O2_name, X_name, theta, K1)(Z1_name);
		let Z2 = this.tkzDefRotaAndScalePoint(O2_name, X_name, -theta, K1)(Z2_name);

		// Y1Z1, Y2Z2 为两条切线，在圆 O1 上的切点为 Y1, Y2，在圆 O2 上的切点为 Z1, Z2
		this.tkz_tmp_points = [{ 'x': Y1.x, 'y': Y1.y }, { 'x': Z1.x, 'y': Z1.y }, { 'x': Y2.x, 'y': Y2.y }, { 'x': Z2.x, 'y': Z2.y }];
		return create_point;
	}

	/**
	 * @brief 作两圆的外公切线，并取得切点。使用方法：tkzDefOutTanCC('O1', 'P1', 'O2', 'P2')('X1,X2,Y1,Y2')，作圆 O1P1 和圆 O2P2 的外公切线 X1X2, Y1Y2，切点为 X1, X2, Y1, Y2
	 * @param O1_name 圆心 1 的名称
	 * @param P1_name 圆 1 上一点名称
	 * @param O2_name 圆心 2 的名称
	 * @param P2_name 圆 2 上一点名称
	 * @return 返回创建点函数，创建四个点。
	 */
	this.tkzDefOutTanCC = function (O1_name, P1_name, O2_name, P2_name) {
		let O1 = this.points[O1_name];
		let O2 = this.points[O2_name];
		let P1 = this.points[P1_name];
		let P2 = this.points[P2_name];

		let O1O2 = this.tkzCalcLength(O1_name, O2_name);
		let r1 = getR(O1, P1);
		let r2 = getR(O2, P2);

		if (Math.abs(r1 - r2) - O1O2 > 0) {
			// 两圆内含，没有外公切线
			this.tkz_msg = '两圆内含，没有外公切线';
			return create_point_err;
		}

		let theta = Math.acos(Math.abs(r1 - r2) / O1O2) * 180 / Math.PI;
		// 将 O1O2 绕 O1 转 theta 到 Y，并缩放到 Y1, Y2，使得 K = O1Y1:Y1Y = r1:(O1O2-r1)
		let K1 = r1 / (O1O2 - r1);

		let [Y1_name, Y2_name] = [UUID.GET(), UUID.GET()];
		let Y1 = this.tkzDefRotaAndScalePoint(O1_name, O2_name, theta, K1)(Y1_name);
		let Y2 = this.tkzDefRotaAndScalePoint(O1_name, O2_name, -theta, K1)(Y2_name);
		// 将 O2O1 绕 O2 转 180-theta 到 Z，并缩放到 Z1, Z2，使得 K = O2Z1:Z1Y = r2:(O2X-r2)
		let K2 = r2 / (O1O2 - r2);
		let [Z1_name, Z2_name] = [UUID.GET(), UUID.GET()];
		let Z1 = this.tkzDefRotaAndScalePoint(O2_name, O1_name, theta - 180, K2)(Z1_name);
		let Z2 = this.tkzDefRotaAndScalePoint(O2_name, O1_name, 180 - theta, K2)(Z2_name);
		// Y1Z1, Y2Z2 为两条切线，在圆 O1 上的切点为 Y1, Y2，在圆 O2 上的切点为 Z1, Z2
		this.tkz_tmp_points = [{ 'x': Y1.x, 'y': Y1.y }, { 'x': Z1.x, 'y': Z1.y }, { 'x': Y2.x, 'y': Y2.y }, { 'x': Z2.x, 'y': Z2.y }];
		return create_point;
	}
	/**
	 * @brief 作点关于圆的反演点。使用方法：tkzDefInvPoint('O', 'A', 'P')('P'')，作点 P 关于圆 OA 的反演点 P'。
	 * @param center_name 圆心点名称
	 * @param point_name 圆上一点名称
	 * @param p_name 被反演的点的名称
	 * @return 返回创建点函数，创建一个点。
	 */
	this.tkzDefInvPoint = function (center_name, point_name, p_name) {
		let p = this.points[p_name];
		let op = this.tkzCalcLength(center_name, p_name);
		// 反演圆半径
		let r = this.tkzCalcLength(center_name, point_name);
		// 定比分点，OQ:QP = K;
		let m = op * op / r / r - 1;
		if (!is_zero(m)) {
			let K = 1 / m;
			return this.tkzDefLinearPoint(center_name, p_name, K);
		} else {
			// 点在圆上，反演点即为自身。
			this.tkz_tmp_points = [{ 'x': p.x, 'y': p.y }];
			return create_point;
		}
	}

	/**
	 * @brief 作调和第四点。使用方法：tkzDefHarmonicPoint('A', 'B', 'C')('D')，作点 D 使得 (AB,CD) = -1。
	 * @param A_name 调和第一点的名称
	 * @param B_name 调和第二点的名称
	 * @param C_name 调和第三点的名称
	 * @return 返回创建点函数，创建一个点。
	 */
	this.tkzDefHarmonicPoint = function (A_name, B_name, C_name) {
		let A = this.points[A_name];
		let B = this.points[B_name];
		let C = this.points[C_name];

		let abc = getABC(A, B);
		if (!is_zero(abc.A * C.x + abc.B * C.y + abc.C)) {
			// 三点不共线，不能创建调和第四点
			this.tkz_msg = '三点不共线，不能创建调和第四点';
			return create_point_err;
		}
		let ac = this.tkzCalcLength(A_name, C_name, true);
		let bc = this.tkzCalcLength(B_name, C_name, true);
		// 调和点列：(AC:BC) = -(AD:BD)
		// 定比分点：K = AD:DB = -(DA:DB) = (AC:BC)
		let K = ac / bc;
		return this.tkzDefLinearPoint(A_name, B_name, K);
	}


	/**
	 * @brief 绘制点
	 * @param params 绘制参数，JSON 格式，例如 {color: 'black', fill: 'black', width: 2} 表示路径为黑色，填充为黑色，点宽为 2 像素
	 * @param points_name 要绘制的点的字符串，以逗号分隔，例如 'A,B'
	 */
	this.tkzDrawPoints = function (params, points_name) {
		params = set_params(params);
		for (let point_name of points_name.split(',')) {
			let point = this.points[point_name];
			ctx.beginPath();
			ctx.strokeStyle = params.color;
			ctx.fillStyle = params.fill;
			ctx.arc(point.x * this.cm_pt, point.y * this.cm_pt, params.width, 0, 2 * Math.PI);
			ctx.fill();
			ctx.stroke();
		}
	}
	/**
	 * @brief 绘制线段
	 * @param params 绘制参数，JSON 格式，例如 {color: 'black', fill: 'black', width: 2, dash: true} 表示路径为黑色，填充为黑色，线宽为 2 像素，虚线
	 * @param start 线段的起点
	 * @param to 线段的终点
	 */
	this.tkzDrawSegment = function (params, start, to) {
		params = set_params(params);
		ctx.beginPath();
		ctx.lineCap = 'butt';
		ctx.lineWidth = params.width;
		ctx.strokeStyle = params.color;
		ctx.moveTo(start.x * this.cm_pt, start.y * this.cm_pt);
		ctx.lineTo(to.x * this.cm_pt, to.y * this.cm_pt);
		ctx.stroke();
	}
	/**
	 * @brief 绘制线段
	 * @param params 绘制参数，JSON 格式，例如 {color: 'black', fill: 'black', width: 2, dash: true} 表示路径为黑色，填充为黑色，线宽为 2 像素，虚线
	 * @param point_pairs_name 线段的起点终点对字符串，起点与终点用逗号分隔，两个线段间用空格分隔，例如 'A,B C,D'
	 */
	this.tkzDrawSegments = function (params, point_pairs_name) {
		for (let point_pairs of point_pairs_name.split(' ')) {
			let start_to = point_pairs.split(',');
			let start = this.points[start_to[0]];
			let to = this.points[start_to[1]];
			this.tkzDrawSegment(params, start, to);
		}
	}
	/**
	 * @brief 绘制圆
	 * @param params 绘制参数，JSON 格式，例如 {color: 'black', width: 2, dash: true} 表示路径为黑色，线宽为 2 像素，虚线
	 * @param center_name 圆心点名称
	 * @param point_name 圆上一点名称
	 */
	this.tkzDrawCircle = function (params, center_name, point_name) {
		params = set_params(params);
		let center = this.points[center_name];
		let point = this.points[point_name];
		ctx.beginPath();
		ctx.lineWidth = params.width;
		ctx.strokeStyle = params.color;
		ctx.arc(center.x * this.cm_pt, center.y * this.cm_pt, getR(center, point) * this.cm_pt, 0, 2 * Math.PI);
		ctx.stroke();
	}

	// let supscript_map = { 'a': 'ᵃ', 'b': 'ᵇ', 'c': 'ᶜ', 'd': 'ᵈ', 'e': 'ᵉ', 'f': 'ᶠ', 'g': 'ᵍ', 'h': 'ʰ', 'i': 'ⁱ', 'j': 'ʲ', 'k': 'ᵏ', 'l': 'ˡ', 'm': 'm', 'n': 'ⁿ', 'o': 'ᵒ', 'p': 'ᵖ', 'r': 'ʳ', 's': 'ˢ', 't': 'ᵗ', 'u': 'ᵘ', 'v': 'ᵛ', 'w': 'ʷ', 'x': 'ˣ', 'y': 'ʸ', 'z': 'ᶻ' };
	let subscript_map = { '0': '₀', '1': '₁', '2': '₂', '3': '₃', '4': '₄', '5': '₅', '6': '₆', '7': '₇', '8': '₈', '9': '₉', 'a': 'ₐ', 'e': 'ₑ', 'i': 'ᵢ', 'o': 'ₒ', 'r': 'ᵣ', 'u': 'ᵤ', 'v': 'ᵥ', 'x': 'ₓ' };
	function getSubNumber(str) {
		if (str.indexOf('_') == -1) {
			return str;
		}
		s = str.match(/(.*)?_{(.*)}/);
		if (s && s.length != 3) {
			return str;
		}
		let new_str = s[1];
		for (let i = 0; i < s[2].length; i++) {
			var code = subscript_map[s[2][i]];
			new_str += code || s[2][i];
		}
		return new_str;
	}
	/**
	 * @brief 画出点名称
	 * @param params 绘画参数
	 * @param point_name 点名称
	 */
	this.tkzLabelPoint = function (params, point_name) {
		params = set_params(params);
		let point = this.points[point_name];
		let p_name = params.label || point.name;
		p_name = getSubNumber(p_name);
		let x = point.x;
		let y = point.y;
		if (typeof (params.r) != 'undefined' && typeof (params.theta) != 'undefined') {
			x += params.r * Math.cos(params.theta / 180 * Math.PI);
			y += params.r * Math.sin(params.theta / 180 * Math.PI);
		}
		if (typeof (params.pos) != 'undefined') {
			params.r = 0.3;
			switch (params.pos) {
				case 'N': params.theta = 90; break;
				case 'S': params.theta = -90; break;
				case 'E': params.theta = 0; break;
				case 'W': params.theta = 180; break;
				case 'EN': params.theta = 45; break;
				case 'ES': params.theta = -45; break;
				case 'WN': params.theta = 135; break;
				case 'WS': params.theta = -135; break;
			}
			x += params.r * Math.cos(params.theta / 180 * Math.PI);
			y += params.r * Math.sin(params.theta / 180 * Math.PI);
		}
		ctx.scale(1, -1);
		ctx.beginPath();
		ctx.fillStyle = params.color;
		ctx.font = params.font;
		ctx.textAlign = 'center';
		ctx.textBaseline = 'middle';
		ctx.fillText(p_name, x * this.cm_pt, -y * this.cm_pt);
		ctx.scale(1, -1);
	}
	/**
	 * @brief 画出点序列名称。使用方法：tkzLabelPoints({color: 'blue'}, 'A,B') 画出 A,B 两点的标记，颜色为 blue
	 * @param params 绘画参数
	 * @param points_name 点序列名称
	 */
	this.tkzLabelPoints = function (params, points_name) {
		points_name = points_name.split(',');
		for (let i = 0; i < points_name.length; i++) {
			this.tkzLabelPoint(params, points_name[i]);
		}
	}
	/**
	 * @brief 画出角名称。使用方法：tkzLabelAngle({color: 'blue', label: 'θ', r: 0.5}, 'A,B,C') 画出角 ABC 的标记，BA 到 BC 为逆时针方向，颜色为 blue，角名称为 θ，角圆弧半径为 0.5
	 * @param params 绘画参数
	 * @param points_name 点序列名称
	 */
	this.tkzLabelAngle = function (params, points_name) {
		points_name = points_name.split(',');
		let A = this.points[points_name[0]];
		let B = this.points[points_name[1]];
		let C = this.points[points_name[2]];
		let start = getDip(B, A) / 180 * Math.PI;
		let to = getDip(B, C) / 180 * Math.PI;
		params.r = params.r || 0.5;
		params.label = params.label || B.name;
		ctx.lineWidth = params.width;
		ctx.strokeStyle = params.color;
		ctx.beginPath();
		ctx.arc(B.x * this.cm_pt, B.y * this.cm_pt, params.r * this.cm_pt, start, to);
		ctx.stroke();
		let len = this.tkzCalcLength(B.name, A.name);
		let K = params.r / (len - params.r - 0.5 / params.r);
		// 画角名称
		let X_name = UUID.GET();
		let theta = this.tkzGetAngle(B.name, A.name, B.name, C.name);
		theta = (theta < 0) ? (360 - theta) : theta;
		params = { color: params.color, width: params.width, label: params.label };
		this.tkzDefRotaAndScalePoint(B.name, A.name, theta / 2, K)(X_name);
		this.tkzLabelPoint(params, X_name);
	}
	/**
	 * @brief 内部方法，创建一个元素。
	 * @param element 元素对象
	 * @param name 元素名称
	 * @return 返回元素对象
	 */
	function create_element(name) {
		it.elements[name] = it.tkz_tmp_elements.pop();
		return it.elements[name];
	}
	/**
	 * @brief 从坐标创建点。使用方法：tkzDefPointsXY([{x:1,y:2},{x:2,y:1}])('A,B')，创建 A(1,2), B(2,1) 两点
	 * @param points_xy 坐标序列，数组格式，每个元素都是 JSON 格式，表示一个坐标
	 * @return 返回创建点函数，创建一组点。
	 */
	this.tkzDefPointsXY = function (points_xy) {
		this.tkz_tmp_points = points_xy;
		return create_point;
	}
	/**
	 * @brief 创建一个椭圆
	 * @param O_name 椭圆中心点名称
	 * @param a 半长轴
	 * @param b 半短轴
	 * @param alpha 倾斜角
	 */
	this.createEllipse = function (O_name, a, b, alpha) {
		alpha = alpha || 0;
		let O = this.points[O_name];
		let c = new tkzEllipse(O.x, O.y, a, b, alpha / 180 * Math.PI);
		this.tkz_tmp_elements.push(c);
		return create_element;
	}
	/**
	 * @brief 创建一个双曲线
	 * @param O_name 双曲线中心点名称
	 * @param a 半实轴
	 * @param b 半虚轴
	 * @param alpha 倾斜角
	 */
	this.createHyperbola = function (O_name, a, b, alpha) {
		alpha = alpha || 0;
		let O = this.points[O_name];
		let c = new tkzHyperbola(O.x, O.y, a, b, alpha / 180 * Math.PI);
		this.tkz_tmp_elements.push(c);
		return create_element;
	}
	/**
	 * @brief 创建一个抛物线
	 * @param O_name 抛物线线顶点名称
	 * @param p 焦准距
	 * @param alpha 倾斜角
	 */
	this.createParabola = function (O_name, p, alpha) {
		alpha = alpha || 0;
		let O = this.points[O_name];
		let c = new tkzParabola(O.x, O.y, p, alpha / 180 * Math.PI);
		this.tkz_tmp_elements.push(c);
		return create_element;
	}
	/**
	 * @brief 创建一条直线
	 * @param A_name 直线上点 1 的名称
	 * @param B_name 直线上点 2 的名称
	 */
	this.createLine = function (A_name, B_name) {
		let A = this.points[A_name];
		let B = this.points[B_name];
		let l = new tkzLine(A.x, A.y, B.x, B.y);
		this.tkz_tmp_elements.push(l);
		return create_element;
	}
	/**
	 * @brief 旋转平移后使得曲线变为标准方程，解方程求得此时直线与椭圆的交点
	 * @param x1 旋转平移后使曲线变为标准方程，此时直线上第一点的 x 坐标
	 * @param y1 旋转平移后使曲线变为标准方程，此时直线上第一点的 y 坐标
	 * @param x2 旋转平移后使曲线变为标准方程，此时直线上第二点的 x 坐标
	 * @param y2 旋转平移后使曲线变为标准方程，此时直线上第二点的 y 坐标
	 * @param c 椭圆对象
	 */
	function solve_ellipse_xy(x1, y1, x2, y2, c) {
		let [xx1, yy1, xx2, yy2] = [0, 0, 0, 0];
		// 旋转平移后的交点坐标
		if (is_zero(x1 - x2)) {
			let p = c.a * c.a - x2 * x2;
			xx1 = x2;
			yy1 = -((c.b * Math.sqrt(p)) / c.a);
			xx2 = x2;
			yy2 = -yy1;
		} else {
			let a2 = c.a * c.a;
			let b2 = c.b * c.b;
			let m = a2 * (y1 - y2);
			let n = b2 * (x1 - x2);
			let d = m * (y1 - y2) + n * (x1 - x2);
			let r = x1 * y2 - x2 * y1;
			let s = a2 * b2 * (x1 - x2) * (x1 - x2) * (d - r * r);
			if (is_zero(s)) {
				s = 0;
			}
			xx1 = (-m * r + Math.sqrt(s)) / d;
			yy1 = (n * (x1 - x2) * r + (y1 - y2) * Math.sqrt(s)) / (d * (x1 - x2));
			xx2 = (-m * r - Math.sqrt(s)) / d;
			yy2 = (n * (x1 - x2) * r - (y1 - y2) * Math.sqrt(s)) / (d * (x1 - x2));
		}
		return [xx1, yy1, xx2, yy2];
	}
	/**
	 * @brief 旋转平移后使得曲线变为标准方程，解方程求得此时直线与双曲线的交点
	 * @param x1 旋转平移后使曲线变为标准方程，此时直线上第一点的 x 坐标
	 * @param y1 旋转平移后使曲线变为标准方程，此时直线上第一点的 y 坐标
	 * @param x2 旋转平移后使曲线变为标准方程，此时直线上第二点的 x 坐标
	 * @param y2 旋转平移后使曲线变为标准方程，此时直线上第二点的 y 坐标
	 * @param c 双曲线对象
	 */
	function solve_hyperbola_xy(x1, y1, x2, y2, c) {
		let [xx1, yy1, xx2, yy2] = [0, 0, 0, 0];
		// 旋转平移后的交点坐标
		if (is_zero(x1 - x2)) {
			let p = x1 * x1 - c.a * c.a;
			xx1 = x1;
			yy1 = -((c.b * Math.sqrt(p)) / c.a);
			xx2 = x1;
			yy2 = -yy1;
		} else {
			let a2 = c.a * c.a;
			let b2 = c.b * c.b;
			let m = a2 * (y1 - y2);
			let n = b2 * (x1 - x2);
			let d = n * (x1 - x2) - m * (y1 - y2);
			let r = x1 * y2 - x2 * y1;
			let s = a2 * b2 * (x1 - x2) * (x1 - x2) * (d + r * r);
			if (is_zero(s)) {
				s = 0;
			}
			xx1 = (m * r + Math.sqrt(s)) / d;
			yy1 = (n * (x1 - x2) * r + (y1 - y2) * Math.sqrt(s)) / (d * (x1 - x2));
			xx2 = (m * r - Math.sqrt(s)) / d;
			yy2 = (n * (x1 - x2) * r - (y1 - y2) * Math.sqrt(s)) / (d * (x1 - x2));
		}
		return [xx1, yy1, xx2, yy2];
	}
	/**
	 * @brief 旋转平移后使得曲线变为标准方程，解方程求得此时直线与双曲线的交点
	 * @param x1 旋转平移后使曲线变为标准方程，此时直线上第一点的 x 坐标
	 * @param y1 旋转平移后使曲线变为标准方程，此时直线上第一点的 y 坐标
	 * @param x2 旋转平移后使曲线变为标准方程，此时直线上第二点的 x 坐标
	 * @param y2 旋转平移后使曲线变为标准方程，此时直线上第二点的 y 坐标
	 * @param c 抛物线对象
	 */
	function solve_parabola_xy(x1, y1, x2, y2, c) {
		let [xx1, yy1, xx2, yy2] = [0, 0, 0, 0];
		// 旋转平移后的交点坐标
		if (is_zero(y1 - y2)) {
			xx1 = y2 * y2 / (2 * c.p);
			yy1 = y2;
			xx2 = xx1;
			yy2 = y2;
		} else {
			let a2 = c.p * (y1 - y2);
			let b2 = c.p * (x1 - x2);
			let n = b2 * (x1 - x2);
			let d = (y1 - y2) * (y1 - y2);
			let r = x1 * y2 - x2 * y1;
			let s = b2 * b2 - 2 * a2 * r;
			if (is_zero(s)) {
				s = 0;
			}
			xx1 = (n - (y1 - y2) * r + x1 * Math.sqrt(s) - x2 * Math.sqrt(s)) / d;
			yy1 = (b2 + Math.sqrt(s)) / (y1 - y2);
			xx2 = (n - (y1 - y2) * r - x1 * Math.sqrt(s) + x2 * Math.sqrt(s)) / d;
			yy2 = (b2 - Math.sqrt(s)) / (y1 - y2);
		}
		return [xx1, yy1, xx2, yy2];
	}
	/**
	 * @brief 取得直线和圆锥曲线的交点
	 * @param l_name 直线的名称
	 * @param c_name 圆锥曲线的名称
	 */
	this.crossLC = function (l_name, c_name) {
		let l = this.elements[l_name];
		let c = this.elements[c_name];
		let [x0, y0, alpha] = [c.x0, c.y0, c.alpha];
		// 将圆锥曲线旋转并平移，变为标准方程，将直线上的两点作同等变换，得到新点 (x1,y1) 和 (x2,y2)
		let [x1, y1, z1] = Matrix.DOT(Matrix.T(-x0, -y0), [l.x1, l.y1, 1]);
		[x1, y1, z1] = Matrix.DOT(Matrix.R(-alpha), [x1, y1, 1]);
		let [x2, y2, z2] = Matrix.DOT(Matrix.T(-x0, -y0), [l.x2, l.y2, 1]);
		[x2, y2, z2] = Matrix.DOT(Matrix.R(-alpha), [x2, y2, 1]);
		let [xx1, yy1, xx2, yy2] = [0, 0, 0, 0];

		switch (c.getType()) {
			case 'hyperbola': [xx1, yy1, xx2, yy2] = solve_hyperbola_xy(x1, y1, x2, y2, c); break;
			case 'parabola': [xx1, yy1, xx2, yy2] = solve_parabola_xy(x1, y1, x2, y2, c); break;
			case 'ellipse': [xx1, yy1, xx2, yy2] = solve_ellipse_xy(x1, y1, x2, y2, c); break;
		}
		let [zz1, zz2] = [0, 0];
		// 旋转平移前的交点坐标
		[xx1, yy1, zz1] = Matrix.DOT(Matrix.R(alpha), [xx1, yy1, 1]);
		[xx1, yy1, zz1] = Matrix.DOT(Matrix.T(x0, y0), [xx1, yy1, 1]);
		[xx2, yy2, zz2] = Matrix.DOT(Matrix.R(alpha), [xx2, yy2, 1]);
		[xx2, yy2, zz2] = Matrix.DOT(Matrix.T(x0, y0), [xx2, yy2, 1]);

		if (is_zero(xx1 - xx2) && is_zero(yy1 - yy2)) {
			// 直线与圆锥曲线相切，只有一个交点
			[xx1, yy1, xx2, yy2] = [l.x1, l.y1, l.x2, l.y2];
		}
		this.tkz_tmp_points = [{ 'x': xx1, 'y': yy1 }, { 'x': xx2, 'y': yy2 }];
		return create_point;
	}
	/**
	 * @brief 取得一点 P 关于曲线 C 的极线
	 * @param P_name 点 P 的名称
	 * @param C_name 曲线 C 的名称
	 * @return 返回极线上两点的坐标
	 */
	this.getPolar = function (P_name, C_name) {
		let p = this.points[P_name];
		let c = this.elements[C_name];
		let v = [p.x, p.y, 1];
		// 取得极线标准方程中的 ABC
		let [A, B, C] = Matrix.DOT(v, c.getMatrix());

		if (is_zero(B)) {
			return [{ 'x': -C / A, 'y': 0 }, { 'x': -C / A, 'y': 1 }];
		}
		if (is_zero(A)) {
			return [{ 'x': 0, 'y': -C / B }, { 'x': 1, 'y': -C / B }];
		}
		if (!is_zero(B) && !is_zero(A)) {
			return [{ 'x': 0, 'y': -C / B }, { 'x': 1, 'y': (-C - A) / B }];
		}
	}
	/**
	 * @brief 取得两点 P1P2 连线关于曲线 C 的极点
	 * @param P1_name 点 P_1 的名称
	 * @param P2_name 点 P_2 的名称
	 * @param C_name 曲线 C 的名称
	 * @return 返回极点的坐标
	 */
	this.getPole = function (P1_name, P2_name, C_name) {
		let P1 = this.points[P1_name];
		let P2 = this.points[P2_name];
		let abc = getABC(P1, P2);
		let v = [abc.A, abc.B, abc.C];
		let c = this.elements[C_name];
		// 取得极点的齐次坐标
		let matrix = c.getMatrix();
		matrix = Matrix.ADJ(matrix);
		let [A, B, C] = Matrix.DOT(v, matrix);

		if (C == 0) {
			return [{ 'x': Infinity, 'y': Infinity }];
		} else {
			return [{ 'x': A / C, 'y': B / C }];
		}
	}
	/**
	 * @brief 取得一点 P 关于曲线 C 的切线
	 * @param P_name 点 P 的名称
	 * @param C_name 曲线 C 的名称
	 * @return 返回创建点函数，创建切点
	 */
	this.getTangent = function (P_name, C_name) {
		// 定义点 P 关于曲线 C 的极线上的两点
		let [A, B] = [UUID.GET(), UUID.GET()];
		this.tkzDefPointsXY(this.getPolar(P_name, C_name))(`${A},${B}`);
		let l = UUID.GET();
		this.createLine(A, B)(l);
		return this.crossLC(l, C_name);
	}
	/**
	 * @brief 画圆锥曲线
	 * @param params 绘制参数，JSON 格式，例如 {color: 'black', fill: 'black', width: 2} 表示路径为黑色，填充为黑色，点宽为 2 像素
	 * @param name 圆锥曲线元素的名称
	 */
	this.drawCurve = function (params, name) {
		params = set_params(params);
		let e = this.elements[name];
		ctx.strokeStyle = params.color;
		for (let t = 0; t < 2 * Math.PI; t += 1E-3) {
			ctx.beginPath();
			let xy = e.getPE(t);
			ctx.arc(xy.x * this.cm_pt, xy.y * this.cm_pt, 0.1, 0, 2 * Math.PI);
			ctx.stroke();
		}
	}
}
